@@ -1,7 +1,7 @@
-/* $Id: subnetcalc.cc 263 2009-06-06 13:37:08Z dreibh $
+/* $Id: subnetcalc.cc 593 2010-06-29 20:48:45Z dreibh $
  *
  * IPv4/IPv6 Subnet Calculator
- * Copyright (C) 2002-2009 by Thomas Dreibholz
+ * Copyright (C) 2002-2010 by Thomas Dreibholz
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,17 +19,25 @@
  * Contact: dreibh@iem.uni-due.de
  */
 
+#include "config.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
-#include <string.h>
+#include <strings.h>
 #include <math.h>
 #include <assert.h>
+#include <locale.h>
+#include <netdb.h>
 #include <iostream>
+#ifdef HAVE_GEOIP
+#include <GeoIP.h>
+#endif
 
 #include "tools.h"
 
 
+
 #ifdef __FreeBSD__   // FreeBSD
 #define s6_addr16 __u6_addr.__u6_addr16
 #define s6_addr32 __u6_addr.__u6_addr32
@@ -205,7 +213,7 @@ void printAddressBinary(std::ostream&         os,
    else {
       int p = 0;
       for(int j = 0;j < 8;j++) {
-         uint16_t a = ntohs(getIPv6Address(address).s6_addr16[j]);
+         uint16_t a = ntohs(((const uint16_t*)&getIPv6Address(address).s6_addr)[j]);
          char str[16];
          snprintf((char*)&str, sizeof(str), "%04x", a);
          os << indent << str << " = ";
@@ -237,7 +245,7 @@ void printAddressBinary(std::ostream&         os,
 
 
 // ###### Is given netmask valid? ###########################################
-unsigned int getPrefixLength(const sockaddr_union& netmask)
+int getPrefixLength(const sockaddr_union& netmask)
 {
    int  prefixLength;
    bool belongsToNetwork = true;
@@ -402,28 +410,28 @@ void printUnicastProperties(std::ostream&   os,
    // ====== Global ID ======================================================
    if(hasGlobalID) {
       char           globalIDString[16];
-      const uint16_t globalID = ntohs(ipv6address.s6_addr16[3]);
+      const uint16_t globalID = ntohs(((const uint16_t*)&ipv6address.s6_addr)[3]);
       snprintf((char*)&globalIDString, sizeof(globalIDString), "%02x%04x%04x",
                ipv6address.s6_addr[1],
-               ntohs(ipv6address.s6_addr16[1]),
-               ntohs(ipv6address.s6_addr16[2]));
+               ntohs(((const uint16_t*)&ipv6address.s6_addr)[1]),
+               ntohs(((const uint16_t*)&ipv6address.s6_addr)[2]));
       os << "      + Global ID    = " << globalIDString << std::endl;
    }
 
    // ====== Subnet ID ======================================================
    if(hasSubnetID) {
       char           subnetIDString[16];
-      const uint16_t subnetID = ntohs(ipv6address.s6_addr16[3]);
+      const uint16_t subnetID = ntohs(((const uint16_t*)&ipv6address.s6_addr)[3]);
       snprintf((char*)&subnetIDString, sizeof(subnetIDString), "%04x", subnetID);
       os << "      + Subnet ID    = " << subnetIDString << std::endl;
    }
 
    // ====== Interface ID ===================================================
    char           interfaceIDString[128];
-   const uint16_t interfaceID[4] = { ntohs(ipv6address.s6_addr16[4]),
-                                     ntohs(ipv6address.s6_addr16[5]),
-                                     ntohs(ipv6address.s6_addr16[6]),
-                                     ntohs(ipv6address.s6_addr16[7]) };
+   const uint16_t interfaceID[4] = { ntohs(((const uint16_t*)&ipv6address.s6_addr)[4]),
+                                     ntohs(((const uint16_t*)&ipv6address.s6_addr)[5]),
+                                     ntohs(((const uint16_t*)&ipv6address.s6_addr)[6]),
+                                     ntohs(((const uint16_t*)&ipv6address.s6_addr)[7]) };
    snprintf((char*)&interfaceIDString, sizeof(interfaceIDString), "\x1b[36m%04x:%02x\x1b[37m%02x:%02x\x1b[38m%02x:%04x\x1b[0m",
             interfaceID[0],
             (interfaceID[1] & 0xff00) >> 8, (interfaceID[1] & 0x00ff),
@@ -447,8 +455,8 @@ void printUnicastProperties(std::ostream&   os,
    char snmcAddressString[32];
    snprintf((char*)&snmcAddressString, sizeof(snmcAddressString),
             "\x1b[32mff02::1:ff\x1b[38m%02x:%04x\x1b[0m",
-            ntohs(ipv6address.s6_addr16[6]) & 0xff,
-            ntohs(ipv6address.s6_addr16[7]));
+            ntohs(((const uint16_t*)&ipv6address.s6_addr)[6]) & 0xff,
+            ntohs(((const uint16_t*)&ipv6address.s6_addr)[7]));
    os << "      + Sol. Node MC = " << snmcAddressString << std::endl;
 }
 
@@ -559,8 +567,8 @@ void printAddressProperties(std::ostream&         os,
 
    // ====== IPv6 properties ================================================
    else {
-      const uint16_t a = ntohs(ipv6address.s6_addr16[0]);
-      const uint16_t b = ntohs(ipv6address.s6_addr16[1]);
+      const uint16_t a = ntohs(((const uint16_t*)&ipv6address.s6_addr)[0]);
+      const uint16_t b = ntohs(((const uint16_t*)&ipv6address.s6_addr)[1]);
 
       // ------ Special addresses -------------------------------------------
       if(IN6_IS_ADDR_LOOPBACK(&ipv6address)) {
@@ -626,13 +634,13 @@ void printAddressProperties(std::ostream&         os,
          // ------ Solicited node multicast address -------------------------
          // FF02::1:FF00:0/104
          if((a == 0xff02) &&
-            (ntohs(ipv6address.s6_addr16[5]) == 0x0001) &&
-            (ntohs(ipv6address.s6_addr16[6]) & 0xff00) == 0xff00) {
+            (ntohs(((const uint16_t*)&ipv6address.s6_addr)[5]) == 0x0001) &&
+            (ntohs(((const uint16_t*)&ipv6address.s6_addr)[6]) & 0xff00) == 0xff00) {
             char nodeAddressString[64];
             snprintf((char*)&nodeAddressString, sizeof(nodeAddressString),
                      "xxxx:xxxx:xxxx:xxxx:xxxx:xxxx:xx%02x:%04x",
-                     ntohs(ipv6address.s6_addr16[6]) >> 8,
-                     ntohs(ipv6address.s6_addr16[7]));
+                     ntohs(((const uint16_t*)&ipv6address.s6_addr)[6]) >> 8,
+                     ntohs(((const uint16_t*)&ipv6address.s6_addr)[7]));
             os << "      + Address is solicited node multicast address for "
                << nodeAddressString << std::endl;
          }
@@ -675,6 +683,7 @@ void printAddressProperties(std::ostream&         os,
 // ###### Main program ######################################################
 int main(int argc, char** argv)
 {
+   bool           noReverseLookup = false;
    int            options;
    int            prefix;
    unsigned int   hostBits;
@@ -689,33 +698,67 @@ int main(int argc, char** argv)
    sockaddr_union host2;
 
 
-   // ====== Get address and netmask ========================================
+   // ====== Initialize locale ==============================================
+   setlocale(LC_ALL, "");   // Necessary to work with IDN names!
+
+   // ====== Print usage ====================================================
    if(argc < 2) {
-      printf("Usage: %s [Address] {Netmask|Prefix} {-uniquelocal|-uniquelocalhq}\n", argv[0]);
+      printf("Usage: %s [Address{/{Netmask|Prefix}}] "
+             "{Netmask|Prefix} {-n} {-uniquelocal|-uniquelocalhq}\n", argv[0]);
       exit(1);
    }
-   if(string2address(argv[1], &address) == false) {
-      printf("ERROR: Bad address %s!\n", argv[1]);
-      exit(1);
-   }
-   if(argc > 2) {
-      options = 3;
-      // ------ Get netmask or prefix ---------------------------------------
-      if(argv[2][0] != '-') {
-         if( (prefix = readPrefix(argv[2], address, netmask)) < 0 ) {
-            if(string2address(argv[2], &netmask) == false) {
-               printf("ERROR: Bad netmask %s!\n", argv[2]);
-               exit(1);
-            }
+
+   // ====== Get address and netmask from one parameter =====================
+   char* slash = index(argv[1], '/');
+   if(slash) {
+      slash[0] = 0x00;
+      if(string2address(argv[1], &address) == false) {
+         printf("ERROR: Bad address %s!\n", argv[1]);
+         exit(1);
+      }
+      if( (prefix = readPrefix((const char*)&slash[1], address, netmask)) < 0 ) {
+         if(string2address((const char*)&slash[1], &netmask) == false) {
+            printf("ERROR: Bad netmask %s!\n", (const char*)&slash[1]);
+            exit(1);
          }
       }
-   }
-   else {
-      // ------ No netmask or prefix => use default for convenience ------
       options = 2;
-      prefix = readPrefix((address.sa.sa_family == AF_INET) ? "32" : "128", address, netmask);
-      assert(prefix >= 0);
    }
+
+   // ====== Get address and netmask from separate parameters ===============
+   else if(slash == NULL) {
+      if(string2address(argv[1], &address) == false) {
+         printf("ERROR: Bad address %s!\n", argv[1]);
+         exit(1);
+      }
+      if(argc > 2) {
+         options = 3;
+         // ------ Get netmask or prefix ------------------------------------
+         if(argv[2][0] != '-') {
+            if( (prefix = readPrefix(argv[2], address, netmask)) < 0 ) {
+               if(string2address(argv[2], &netmask) == false) {
+                  printf("ERROR: Bad netmask %s!\n", argv[2]);
+                  exit(1);
+               }
+            }
+         }
+         else {
+            // ------ No netmask or prefix => use default for convenience ---
+            options = 2;
+            prefix = readPrefix((address.sa.sa_family == AF_INET) ? "32" : "128", address, netmask);
+            assert(prefix >= 0);
+         }
+      }
+      else {
+         // ------ No netmask or prefix => use default for convenience ------
+         options = 2;
+         prefix = readPrefix((address.sa.sa_family == AF_INET) ? "32" : "128", address, netmask);
+         assert(prefix >= 0);
+      }
+   }
+
+
+   // ====== Get prefix length ==============================================
    prefix = getPrefixLength(netmask);
    if(prefix < 0) {
       std::cerr << "ERROR: Invalid netmask ";
@@ -739,6 +782,9 @@ int main(int argc, char** argv)
       else if(strcmp(argv[i], "-uniquelocalhq") == 0) {
          generateUniqueLocal(address, true);
       }
+      else if(strcmp(argv[i], "-n") == 0) {
+         noReverseLookup = true;
+      }
       else {
          std::cerr << "ERROR: Bad argument " << argv[i] << "!" << std::endl;
          exit(1);
@@ -792,5 +838,68 @@ int main(int argc, char** argv)
       std::cout << "Max. Hosts    = " << maxHostsString << std::endl;
       std::cout << "Host Range    = { " << host1 << " - " << host2 << " }" << std::endl;
    }
+
+
+   // ====== Properties =====================================================
    printAddressProperties(std::cout, address, netmask, prefix, network, broadcast);
+
+
+   // ====== GeoIP ==========================================================
+#ifdef HAVE_GEOIP
+   const char* country = NULL;
+   const char* code    = NULL;
+   if(address.sa.sa_family == AF_INET) {
+      GeoIP* geoIP = GeoIP_open_type(GEOIP_COUNTRY_EDITION, GEOIP_STANDARD);
+      if(geoIP) {
+         country = GeoIP_country_name_by_ipnum(geoIP, ntohl(address.in.sin_addr.s_addr));
+         code    = GeoIP_country_code_by_ipnum(geoIP, ntohl(address.in.sin_addr.s_addr));
+         std::cout << "GeoIP Country = "
+                   << ((country != NULL) ? country: "Unknown")
+                   << " (" << ((code != NULL) ? code : "??") << ")" << std::endl;
+         GeoIP_delete(geoIP);
+      }
+   }
+#ifdef HAVE_GEOIP_IPV6
+   else if(address.sa.sa_family == AF_INET6) {
+      GeoIP* geoIP = GeoIP_open_type(GEOIP_COUNTRY_EDITION_V6, GEOIP_STANDARD);
+      if(geoIP) {
+         country = GeoIP_country_name_by_ipnum_v6(geoIP, address.in6.sin6_addr);
+         code    = GeoIP_country_code_by_ipnum_v6(geoIP, address.in6.sin6_addr);
+         std::cout << "GeoIP Country = "
+                   << ((country != NULL) ? country: "Unknown")
+                   << " (" << ((code != NULL) ? code : "??") << ")" << std::endl;
+         GeoIP_delete(geoIP);
+      }
+   }
+#else
+#warning This used version of GeoIP does not yet support IPv6!
+#endif
+#endif
+
+
+   // ====== Reverse lookup =================================================
+   if(noReverseLookup == false) {
+      std::cout << "Performing reverse DNS lookup ..."; std::cout.flush();
+      char hostname[NI_MAXHOST];
+      int error = getnameinfo(&address.sa,
+                              (address.sa.sa_family == AF_INET6) ?
+                                 sizeof(sockaddr_in6) : sizeof(sockaddr_in),
+                              (char*)&hostname, sizeof(hostname),
+                              NULL, 0,
+#ifdef NI_IDN
+                              NI_NAMEREQD|NI_IDN
+#else
+#warning No IDN support for getnameinfo()!
+                              NI_NAMEREQD
+#endif
+                              );
+      std::cout << "\r\x1b[K";
+      std::cout << "DNS Hostname  = "; std::cout.flush();
+      if(error == 0) {
+         std::cout << hostname << std::endl;
+      }
+      else {
+         std::cout << "(" << gai_strerror(error) << ")" << std::endl;
+      }
+   }
 }
